import multiprocessing
import os
import re
import socket
import json
import tomli
from distutils.sysconfig import get_python_lib
from maxwell.utils.logger import get_logger

logger = get_logger(__name__)


class Config:
    __instance = None

    # ===========================================
    # apis
    # ===========================================
    def __init__(self):
        self.__service_config = self.__build_service_config()
        self.__log_config = self.__build_log_config()

    @staticmethod
    def singleton():
        if Config.__instance == None:
            Config.__instance = Config()
        return Config.__instance

    def get_port(self):
        port = self.__service_config.get("port")
        if port is None:
            port = self.__get_unused_port()
            self.__service_config["port"] = port
            self.__save_port_to_config_file(port)
        return port

    def get_workers(self):
        workers = self.__service_config.get("workers")
        if workers is None or workers == -1:
            return multiprocessing.cpu_count() * 2 + 1
        else:
            return workers

    def get_proc_name(self):
        proc_name = self.__service_config.get("proc_name")
        if proc_name is None:
            return "maxwell-service-python"
        else:
            return proc_name

    def get_master_endpoints(self):
        master_endpoints = self.__service_config.get("master_endpoints")
        if master_endpoints is None:
            raise "Please specify master_endpoints in service.toml"
        else:
            return master_endpoints

    def get_connection_slot_size(self):
        connection_slot_size = self.__service_config.get("connection_slot_size")
        if connection_slot_size is None:
            return 8
        else:
            return connection_slot_size

    def get_endpoint_cache_size(self):
        endpoint_cache_size = self.__service_config.get("endpoint_cache_size")
        if endpoint_cache_size is None:
            return 20480
        else:
            return endpoint_cache_size

    def get_endpoint_cache_ttl(self):
        endpoint_cache_ttl = self.__service_config.get("endpoint_cache_ttl")
        if endpoint_cache_ttl is None:
            return 60 * 60 * 24
        else:
            return endpoint_cache_ttl

    def get_max_continuous_disconnected_times(self):
        max_continuous_disconnected_times = self.__service_config.get(
            "max_continuous_disconnected_times"
        )
        if max_continuous_disconnected_times is None:
            return 30
        else:
            return max_continuous_disconnected_times

    def get_log_config(self):
        return self.__log_config

    # ===========================================
    # internal functions
    # ===========================================
    def __build_service_config(self):
        config_path = self.__get_service_config_file()
        if os.path.exists(config_path):
            with open(config_path, "rb") as f:
                return tomli.load(f)
        else:
            return {}

    def __get_service_config_file(self):
        specified_config_file = os.getenv("SERVICE_CFG_FILE", None)
        if specified_config_file:
            config_file = specified_config_file
        else:
            config_file = os.path.join(self.__get_root_dir(), "config", "service.toml")
        return config_file

    def __build_log_config(self):
        config_path = self.__get_log_config_file()
        if os.path.exists(config_path):
            with open(config_path, "rt") as config_file:
                config = json.load(config_file)
                log_dir = self.__get_log_dir()
                for handler in config["handlers"].values():
                    if "filename" in handler:
                        handler["filename"] = os.path.join(log_dir, handler["filename"])
                return config
        else:
            return None

    def __get_log_config_file(self):
        specified_config_file = os.getenv("LOG_CFG_FILE", None)
        if specified_config_file:
            config_file = specified_config_file
        else:
            config_file = os.path.join(self.__get_root_dir(), "config", "logging.json")
        return config_file

    def __get_root_dir(self):
        root_dir = os.path.abspath(
            os.path.join(get_python_lib(), "..", "..", "..", "..")
        )
        return root_dir

    def __get_log_dir(self):
        return os.path.join(self.__get_root_dir(), "log")

    def __get_unused_port(self):
        # Implement this function instead of using portpicker to be compatiable with proxy chains
        attempt = 1024
        while attempt > 0:
            try:
                with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
                    s.bind(("localhost", 0))
                    _, port = s.getsockname()
                    return port
            except Exception as e:
                logger.warning("Error occurred: %s", e)
                attempt -= 1
        raise LookupError("Failed to find unused port.")

    def __save_port_to_config_file(self, port):
        try:
            config_path = self.__get_service_config_file()
            append_string = f"port = {port}\n"
            with open(config_path, "r+") as file:
                content = file.read()
                if not re.search(r"^port", content, re.MULTILINE):
                    file.write(append_string + "\n")
                    logger.info(f"Writed new port: %s", port)
                else:
                    logger.info(f"Ignored, as port already exists: %s", port)
        except Exception as e:
            logger.warning("Error occurred: %s", e)



import asyncio
import json
import threading
from maxwell.utils.logger import get_logger
from maxwell.service.server import Server
from maxwell.service.service import Service
from maxwell.service.publisher import Publisher

logger = get_logger(__name__)
service = Service()


@service.ws("/hello")
async def hello(req):
    logger.debug(" %s ", req)
    return json.dumps("world")


@service.ws("/get_candles")
async def hello(req):
    logger.debug(" %s ", req)
    return json.dumps(build_candles())


def build_candles():
    candles = []
    for i in range(0, 30000):
        candles.append(
            {
                "ts": i,
                "open": i + 1,
                "high": i + 2,
                "low": i + 3,
                "close": i + 4,
                "volume": i + 5,
            }
        )
    return candles


def run_publisher(loop):
    loop.run_until_complete(run_publisher_coro(loop))


async def run_publisher_coro(loop):
    publisher = Publisher(options={}, loop=loop)
    while True:
        try:
            rep = await publisher.publish("topic_1", b"hello world")
            logger.info("***published: %s", rep)
        except Exception as e:
            logger.error("Failed to publish: %s", e)

        await asyncio.sleep(1)


if __name__ == "__main__":
    loop = asyncio.new_event_loop()
    t = threading.Thread(target=run_publisher, args=(loop,), daemon=True)
    t.start()
    Server(service.register()).run()



import asyncio
from async_lru import alru_cache
import maxwell.protocol.maxwell_protocol_pb2 as protocol_types
from maxwell.utils.connection import Event
from maxwell.utils.logger import get_logger

from .config import Config
from .master_client import MasterClient

logger = get_logger(__name__)


class TopicLocatlizer(object):
    # ===========================================
    # apis
    # ===========================================
    def __init__(self, loop):
        self.__loop = loop

        self.__checksum = 0
        self.__master_client = MasterClient(
            Config.singleton().get_master_endpoints(),
            {"reconnect_delay": 1, "ping_interval": 10},
            self.__loop,
        )
        self.__master_client.add_connection_listener(
            Event.ON_CONNECTED, self.__on_connected_to_master
        )

    def __del__(self):
        asyncio.ensure_future(self.close())

    async def close(self):
        self.__master_client.delete_connection_listener(
            Event.ON_CONNECTED, self.__on_connected_to_master
        )
        await self.__master_client.close()
        self.locate.cache_clear()
        await self.locate.cache_close()

    @alru_cache(
        maxsize=Config.singleton().get_endpoint_cache_size(),
        ttl=Config.singleton().get_endpoint_cache_ttl(),
    )
    async def locate(self, topic):
        req = protocol_types.locate_topic_req_t()
        req.topic = topic
        rep = await self.__master_client.request(req)
        return rep.endpoint

    # ===========================================
    # internal functions
    # ===========================================
    def __on_connected_to_master(self, *argv, **kwargs):
        self.__loop.create_task(self.__check())

    async def __check(self):
        req = protocol_types.get_topic_dist_checksum_req_t()
        logger.info("Getting TopicDistChecksum: req: %s", req)
        try:
            rep = await self.__master_client.request(req)
            logger.info("Successfully to get TopicDistChecksum: rep: %s", rep)
        except Exception as e:
            logger.error("Failed to get TopicDistChecksum: %s", e)
            return

        if self.__checksum != rep.checksum:
            logger.info(
                "TopicDistChecksum has changed: local: %s, remote: %s, clear cache...",
                self.__checksum,
                rep.checksum,
            )
            self.__checksum = rep.checksum
            self.locate.cache_clear()
        else:
            logger.info(
                "TopicDistChecksum stays the same: local: %s, remote: %s, do nothing.",
                self.__checksum,
                rep.checksum,
            )



import asyncio
import functools
import inspect
import traceback
from typing import Any
from fastapi import FastAPI, WebSocket, WebSocketDisconnect
from maxwell.utils.logger import get_logger
import maxwell.protocol.maxwell_protocol_pb2 as protocol_types
import maxwell.protocol.maxwell_protocol as protocol

logger = get_logger(__name__)


class Service(FastAPI):
    def __init__(self, *args: Any, **kwargs: Any):
        super(Service, self).__init__(*args, **kwargs)
        self.__is_registered = False
        self.__ws_routes = {}
        self.__init_ws()

    def register(self):
        if self.__is_registered is False:
            self.__is_registered = True
        return self

    def is_registered(self):
        return self.__is_registered

    def ws(self, path):
        def decorator(func):
            @functools.wraps(func)
            def func_wrapper(*args, **kwargs):
                value = func(*args, **kwargs)
                return value

            self.__ws_routes[path] = [func_wrapper, inspect.iscoroutinefunction(func)]
            return func_wrapper

        return decorator

    def get_ws_routes(self):
        return self.__ws_routes

    def get_paths(self):
        return list(self.__ws_routes.keys())

    def get_handler(self, path):
        return self.__ws_routes.get(path)

    def __init_ws(self):
        @self.websocket("/$ws")
        async def websocket_endpoint(websocket: WebSocket):
            await websocket.accept()
            try:
                while True:
                    data = await websocket.receive_bytes()
                    asyncio.ensure_future(self.__handle_msg(websocket, data))
            except WebSocketDisconnect as e:
                logger.warning("Connection was closed: %s", e)

    async def __handle_msg(self, websocket, data):
        try:
            req = protocol.decode_msg(data)
            if req.__class__ == protocol_types.req_req_t:
                handler = self.get_handler(req.path)
                if handler is not None:
                    logger.debug("Received msg: %s", req)
                    rep = protocol_types.req_rep_t()
                    if handler[1] is True:
                        rep.payload = await handler[0](req)
                    else:
                        rep.payload = handler[0](req)
                    rep.conn0_ref = req.conn0_ref
                    rep.ref = req.ref
                    await websocket.send_bytes(protocol.encode_msg(rep))
                else:
                    logger.error("Unknown path: %s", req.path)
                    rep = protocol_types.error2_rep_t()
                    rep.code = protocol_types.error_code_t.UNKNOWN_PATH
                    rep.desc = "Unknown path: %s" % req.path
                    rep.conn0_ref = req.conn0_ref
                    rep.ref = req.ref
                    await websocket.send_bytes(protocol.encode_msg(rep))
            elif req.__class__ == protocol_types.ping_req_t:
                rep = protocol_types.ping_rep_t()
                rep.ref = req.ref
                await websocket.send_bytes(protocol.encode_msg(rep))
            else:
                logger.error("Received unknown msg: %s", req)
        except Exception:
            logger.error(
                "Failed to handle msg: %s, error: %s", req, traceback.format_exc()
            )



from typing import Any, Dict
from uvicorn.workers import UvicornWorker
from .config import Config


class Worker(UvicornWorker):
    CONFIG_KWARGS: Dict[str, Any] = {
        "loop": "uvloop",
        "http": "httptools",
        "ws": "websockets",
        "ws_max_size": 134217728,
        "ws_ping_interval": None,
        "ws_ping_timeout": None,
        "lifespan": "on",
        "interface": "asgi3",
        "log_config": Config.singleton().get_log_config(),
    }

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super(Worker, self).__init__(*args, **kwargs)



from maxwell.service.config import Config


class TestConfig:
    def test_all(self):
        config = Config.singleton()
        assert config.get_port() == 9091
        assert config.get_workers() == 2
        assert config.get_proc_name() == "maxwell-service-python"



import asyncio
import random
from maxwell.utils.connection import Connection, Event
from maxwell.utils.logger import get_logger
import maxwell.protocol.maxwell_protocol_pb2 as protocol_types

from .config import Config
from .topic_locatlizer import TopicLocatlizer

logger = get_logger(__name__)


class Publisher(object):
    # ===========================================
    # apis
    # ===========================================
    def __init__(self, options, loop):
        self.__options = options
        self.__loop = loop

        self.__topic_locatlizer = TopicLocatlizer(self.__loop)
        self.__connections = {}  # endpoint => [connection0, connection1, ...]
        self.__continuous_disconnected_times = 0

    def __del__(self):
        asyncio.ensure_future(self.close())

    async def close(self):
        for connections in self.__connections.values():
            for connection in connections:
                await connection.close()
        await self.__topic_locatlizer.close()

    async def publish(self, topic, value):
        try:
            connection = await self.__get_connetion(topic)
            await connection.wait_open()
            return await connection.request(self.__build_publish_req(topic, value))
        except Exception as e:
            logger.error("Failed to publish: topic: %s, error: %s", topic, e)

    # ===========================================
    # internal functions
    # ===========================================
    async def __get_connetion(self, topic):
        endpoint = await self.__topic_locatlizer.locate(topic)
        connections = self.__connections.get(endpoint)
        size = Config.singleton().get_connection_slot_size()
        if connections is None:
            connections = []
            for _ in range(size):
                connection = Connection(
                    endpoint=endpoint, options=self.__options, loop=self.__loop
                )
                connection.add_listener(
                    event=Event.ON_DISCONNECTED,
                    callback=self.__on_disconnected_to_backend,
                )
                connections.append(connection)
            self.__connections[endpoint] = connections
        return connections[random.randint(0, size - 1)]

    def __on_disconnected_to_backend(self, connection):
        self.__continuous_disconnected_times += 1
        if (
            self.__continuous_disconnected_times
            >= Config.singleton().get_max_continuous_disconnected_times()
        ):
            self.__continuous_disconnected_times = 0
            self.__connections.pop(connection.get_endpoint(), None)

    def __build_publish_req(self, topic, value):
        push_req = protocol_types.push_req_t()
        push_req.topic = topic
        push_req.value = value
        return push_req



class Hooks(object):
    # post_service_init(worker): Called just after a worker has initialized the application.
    # post_worker_fork(server, worker): Called just after a worker has been forked.
    # post_worker_exit(server, worker): Called just after a worker has been exited, in the worker process.
    def __init__(
        self, post_service_init=None, post_worker_fork=None, post_worker_exit=None
    ):
        super().__init__()
        self.post_service_init = (
            post_service_init if post_service_init else self.__default_post_service_init
        )
        self.post_worker_fork = (
            post_worker_fork if post_worker_fork else self.__default_post_worker_fork
        )
        self.post_worker_exit = (
            post_worker_exit if post_worker_exit else self.__default_post_worker_exit
        )

    def __default_post_service_init(self, worker):
        pass

    def __default_post_worker_fork(self, server, worker):
        pass

    def __default_post_worker_exit(self, server, worker):
        pass



import asyncio
import threading
import maxwell.protocol.maxwell_protocol_pb2 as protocol_types
from maxwell.utils.connection import Event
from maxwell.utils.logger import get_logger

from .config import Config
from .master_client import MasterClient

logger = get_logger(__name__)


class Registrar(object):
    # ===========================================
    # apis
    # ===========================================
    def __init__(self, queue, loop=None):
        self.__queue = queue
        self.__loop = loop if loop else asyncio.new_event_loop()

        self.__running = True
        self.__should_register_event = asyncio.Event()
        self.__should_register_event.clear()

    def start(self):
        t = threading.Thread(target=self.__run_loop, args=(), daemon=True)
        t.start()
        logger.info("Registrar thread started.")

    def stop(self):
        self.__running = False
        self.__should_register_event.clear()

    # ===========================================
    # internal functions
    # ===========================================
    def __run_loop(self):
        self.__loop.run_until_complete(self.__repeat_register())
        logger.info("Registrar thread stopped.")

    async def __repeat_register(self):
        paths = self.__queue.get()
        logger.info("Received paths from worker: %s", paths)

        master_client = MasterClient(
            Config.singleton().get_master_endpoints(),
            {"reconnect_delay": 1, "ping_interval": 10},
            self.__loop,
        )
        master_client.add_connection_listener(
            event=Event.ON_CONNECTED, callback=self.__on_connected_to_master
        )

        while self.__running:
            try:
                await self.__should_register_event.wait()
                await self.__register_service(master_client)
                await self.__set_routes(paths, master_client)
                self.__should_register_event.clear()
            except Exception as e:
                logger.error("Failed to register: %s", e)
                await asyncio.sleep(1)

        master_client.delete_connection_listener(
            event=Event.ON_CONNECTED, callback=self.__on_connected_to_master
        )
        await master_client.close()

    def __on_connected_to_master(self, *argv, **kwargs):
        self.__should_register_event.set()

    async def __register_service(self, master_client):
        req = protocol_types.register_service_req_t()
        req.http_port = Config.singleton().get_port()
        try:
            rep = await master_client.request(req)
            logger.info("Successfully to register service: %s", rep)
        except Exception as e:
            logger.error("Failed to register service: %s", e)
            raise e

    async def __set_routes(self, paths, master_client):
        req = protocol_types.set_routes_req_t()
        req.paths.extend(paths)
        try:
            rep = await master_client.request(req)
            logger.info("successfully to set routes: %s", rep)
        except Exception as e:
            logger.error("Failed to set routes: %s", e)
            raise e



import threading
import time
import gunicorn.app.base
from ctypes import c_bool
from multiprocessing import Queue, Value
from maxwell.utils.logger import get_logger

from .config import Config
from .hooks import Hooks
from .registrar import Registrar

logger = get_logger(__name__)


class Server(gunicorn.app.base.BaseApplication):
    def __init__(self, service, hooks=None):
        config = Config.singleton()
        hooks = hooks or Hooks()

        self.options = {
            "bind": ["0.0.0.0:{}".format(config.get_port())],
            "workers": config.get_workers(),
            "proc_name": config.get_proc_name(),
            "logconfig_dict": config.get_log_config(),
            "worker_class": "maxwell.service.worker.Worker",
            "proxy_allow_ips": "*",
            "post_worker_init": self.__post_worker_init,
            "post_fork": self.__post_fork,
            "worker_exit": self.__worker_exit,
            "when_ready": self.__when_ready,
            "on_exit": self.__on_exit,
        }
        self.application = service
        super().__init__()

        self.__service = service
        self.__hooks = hooks
        self.__queue = Queue()
        self.__is_first_registered_service = Value(c_bool, True)
        self.__registrar = None

    def load_config(self):
        config = {
            key: value
            for key, value in self.options.items()
            if key in self.cfg.settings and value is not None
        }
        for key, value in config.items():
            self.cfg.set(key.lower(), value)

    def load(self):
        return self.application

    def __post_worker_init(self, worker):
        def wait_service_to_register():
            while True:
                if self.__service.is_registered():
                    logger.info("[2] post_service_init: worker: %s", worker)
                    self.__hooks.post_service_init(worker)

                    with self.__is_first_registered_service.get_lock():
                        if self.__is_first_registered_service.value is True:
                            self.__is_first_registered_service.value = False
                            paths = self.__service.get_paths()
                            logger.info("Sending paths to registrar: %s", paths)
                            self.__queue.put(paths)

                    break

                time.sleep(0.1)

        t = threading.Thread(target=wait_service_to_register, args=(), daemon=True)
        t.start()

    def __post_fork(self, server, worker):
        logger.info("[1] post_worker_fork: server: %s, worker: %s", server, worker)
        self.__hooks.post_worker_fork(server, worker)

    def __worker_exit(self, server, worker):
        logger.info("[3] post_worker_exit: server: %s, worker: %s", server, worker)
        self.__hooks.post_worker_exit(server, worker)

    def __when_ready(self, server):
        logger.info("[0] Server started: server: %s", server)
        if self.__registrar is None:
            self.__registrar = Registrar(queue=self.__queue)
            self.__registrar.start()

    def __on_exit(self, server):
        logger.info("[N] Server exit: server: %s", server)
        if self.__registrar is not None:
            self.__registrar.stop()



import asyncio
from maxwell.utils.connection import MultiAltEndpointsConnection


class MasterClient(object):
    __instance = None

    # ===========================================
    # apis
    # ===========================================
    def __init__(self, endpoints, options, loop):
        self.__endpoints = endpoints
        self.__options = options
        self.__loop = loop

        self.__endpoint_index = -1
        self.__connection = MultiAltEndpointsConnection(
            pick_endpoint=self.__pick_endpoint,
            options=self.__options,
            loop=self.__loop,
        )

    def __del__(self):
        asyncio.ensure_future(self.close())

    @staticmethod
    def singleton(endpoints, options, loop):
        if MasterClient.__instance == None:
            MasterClient.__instance = MasterClient(endpoints, options, loop)
        return MasterClient.__instance

    async def close(self):
        await self.__connection.close()
        self.__connection = None

    def add_connection_listener(self, event, callback):
        self.__connection.add_listener(event, callback)

    def delete_connection_listener(self, event, callback):
        self.__connection.delete_listener(event, callback)

    async def request(self, msg):
        await self.__connection.wait_open()
        return await self.__connection.request(msg)

    # ===========================================
    # internal functions
    # ===========================================
    async def __pick_endpoint(self):
        self.__endpoint_index += 1
        if self.__endpoint_index >= len(self.__endpoints):
            self.__endpoint_index = 0
        return self.__endpoints[self.__endpoint_index]
